static const char* const ip_protocol_names[] = { 
        [IPPROTO_IP] = "ip",
        [IPPROTO_ROUTING] = "routing",
        [IPPROTO_IDP] = "idp",
        [IPPROTO_SCTP] = "sctp",
        [IPPROTO_EGP] = "egp",
        [IPPROTO_PIM] = "pim",
        [IPPROTO_ENCAP] = "encap",
        [IPPROTO_ESP] = "esp",
        [IPPROTO_BEETPH] = "beetph",
        [IPPROTO_PUP] = "pup",
        [IPPROTO_IPIP] = "ipip",
        [IPPROTO_TCP] = "tcp",
        [IPPROTO_RAW] = "raw",
        [IPPROTO_UDP] = "udp",
        [IPPROTO_TP] = "tp",
        [IPPROTO_IPV6] = "ipv6",
        [IPPROTO_RSVP] = "rsvp",
        [IPPROTO_AH] = "ah",
        [IPPROTO_ICMP] = "icmp",
        [IPPROTO_MH] = "mh",
        [IPPROTO_MTP] = "mtp",
        [IPPROTO_ICMPV6] = "icmpv6",
        [IPPROTO_MPLS] = "mpls",
        [IPPROTO_MPTCP] = "mptcp",
        [IPPROTO_GRE] = "gre",
        [IPPROTO_DSTOPTS] = "dstopts",
        [IPPROTO_NONE] = "none",
        [IPPROTO_IGMP] = "igmp",
        [IPPROTO_COMP] = "comp",
        [IPPROTO_DCCP] = "dccp",
        [IPPROTO_FRAGMENT] = "fragment",
        [IPPROTO_ETHERNET] = "ethernet",
        [IPPROTO_UDPLITE] = "udplite",
};
