case (statfs_f_type_t) AAFS_MAGIC:
        return "apparmorfs";
case (statfs_f_type_t) ADFS_SUPER_MAGIC:
        return "adfs";
case (statfs_f_type_t) AFFS_SUPER_MAGIC:
        return "affs";
case (statfs_f_type_t) AFS_FS_MAGIC:
case (statfs_f_type_t) AFS_SUPER_MAGIC:
        return "afs";
case (statfs_f_type_t) ANON_INODE_FS_MAGIC:
        return "anon_inodefs";
case (statfs_f_type_t) AUTOFS_SUPER_MAGIC:
        return "autofs";
case (statfs_f_type_t) BALLOON_KVM_MAGIC:
        return "balloon-kvm";
case (statfs_f_type_t) BDEVFS_MAGIC:
        return "bdev";
case (statfs_f_type_t) BINDERFS_SUPER_MAGIC:
        return "binder";
case (statfs_f_type_t) BINFMTFS_MAGIC:
        return "binfmt_misc";
case (statfs_f_type_t) BPF_FS_MAGIC:
        return "bpf";
case (statfs_f_type_t) BTRFS_SUPER_MAGIC:
        return "btrfs";
case (statfs_f_type_t) BTRFS_TEST_MAGIC:
        return "btrfs_test_fs";
case (statfs_f_type_t) CEPH_SUPER_MAGIC:
        return "ceph";
case (statfs_f_type_t) CGROUP2_SUPER_MAGIC:
        return "cgroup2";
case (statfs_f_type_t) CGROUP_SUPER_MAGIC:
        return "cgroup";
case (statfs_f_type_t) CIFS_SUPER_MAGIC:
case (statfs_f_type_t) SMB2_SUPER_MAGIC:
        return "cifs";
case (statfs_f_type_t) CODA_SUPER_MAGIC:
        return "coda";
case (statfs_f_type_t) CONFIGFS_MAGIC:
        return "configfs";
case (statfs_f_type_t) CRAMFS_MAGIC:
        return "cramfs";
case (statfs_f_type_t) DAXFS_MAGIC:
        return "dax";
case (statfs_f_type_t) DEBUGFS_MAGIC:
        return "debugfs";
case (statfs_f_type_t) DEVMEM_MAGIC:
        return "devmem";
case (statfs_f_type_t) DEVPTS_SUPER_MAGIC:
        return "devpts";
case (statfs_f_type_t) DMA_BUF_MAGIC:
        return "dmabuf";
case (statfs_f_type_t) ECRYPTFS_SUPER_MAGIC:
        return "ecryptfs";
case (statfs_f_type_t) EFIVARFS_MAGIC:
        return "efivarfs";
case (statfs_f_type_t) EFS_SUPER_MAGIC:
        return "efs";
case (statfs_f_type_t) EROFS_SUPER_MAGIC_V1:
        return "erofs";
case (statfs_f_type_t) EXT4_SUPER_MAGIC:
        return "ext4";
case (statfs_f_type_t) EXFAT_SUPER_MAGIC:
        return "exfat";
case (statfs_f_type_t) F2FS_SUPER_MAGIC:
        return "f2fs";
case (statfs_f_type_t) FUSE_SUPER_MAGIC:
        return "fuse";
case (statfs_f_type_t) FUSE_CTL_SUPER_MAGIC:
        return "fusectl";
case (statfs_f_type_t) GFS2_MAGIC:
        return "gfs2";
case (statfs_f_type_t) HOSTFS_SUPER_MAGIC:
        return "hostfs";
case (statfs_f_type_t) HPFS_SUPER_MAGIC:
        return "hpfs";
case (statfs_f_type_t) HUGETLBFS_MAGIC:
        return "hugetlbfs";
case (statfs_f_type_t) ISOFS_SUPER_MAGIC:
        return "iso9660";
case (statfs_f_type_t) JFFS2_SUPER_MAGIC:
        return "jffs2";
case (statfs_f_type_t) MINIX_SUPER_MAGIC:
case (statfs_f_type_t) MINIX_SUPER_MAGIC2:
case (statfs_f_type_t) MINIX2_SUPER_MAGIC:
case (statfs_f_type_t) MINIX2_SUPER_MAGIC2:
case (statfs_f_type_t) MINIX3_SUPER_MAGIC:
        return "minix";
case (statfs_f_type_t) MQUEUE_MAGIC:
        return "mqueue";
case (statfs_f_type_t) NCP_SUPER_MAGIC:
        return "ncpfs";
case (statfs_f_type_t) NFS_SUPER_MAGIC:
        return "nfs4";
case (statfs_f_type_t) NILFS_SUPER_MAGIC:
        return "nilfs2";
case (statfs_f_type_t) NSFS_MAGIC:
        return "nsfs";
case (statfs_f_type_t) NTFS_SB_MAGIC:
        return "ntfs";
case (statfs_f_type_t) NTFS3_SUPER_MAGIC:
        return "ntfs3";
case (statfs_f_type_t) OCFS2_SUPER_MAGIC:
        return "ocfs2";
case (statfs_f_type_t) OPENPROM_SUPER_MAGIC:
        return "openpromfs";
case (statfs_f_type_t) ORANGEFS_DEVREQ_MAGIC:
        return "orangefs";
case (statfs_f_type_t) OVERLAYFS_SUPER_MAGIC:
        return "overlay";
case (statfs_f_type_t) PIPEFS_MAGIC:
        return "pipefs";
case (statfs_f_type_t) PPC_CMM_MAGIC:
        return "ppc-cmm";
case (statfs_f_type_t) PROC_SUPER_MAGIC:
        return "proc";
case (statfs_f_type_t) PSTOREFS_MAGIC:
        return "pstore";
case (statfs_f_type_t) QNX4_SUPER_MAGIC:
        return "qnx4";
case (statfs_f_type_t) QNX6_SUPER_MAGIC:
        return "qnx6";
case (statfs_f_type_t) RAMFS_MAGIC:
        return "ramfs";
case (statfs_f_type_t) RDTGROUP_SUPER_MAGIC:
        return "resctrl";
case (statfs_f_type_t) REISERFS_SUPER_MAGIC:
        return "reiserfs";
case (statfs_f_type_t) RPC_PIPEFS_SUPER_MAGIC:
        return "rpc_pipefs";
case (statfs_f_type_t) SECRETMEM_MAGIC:
        return "secretmem";
case (statfs_f_type_t) SECURITYFS_MAGIC:
        return "securityfs";
case (statfs_f_type_t) SELINUX_MAGIC:
        return "selinuxfs";
case (statfs_f_type_t) SHIFTFS_MAGIC:
        return "shiftfs";
case (statfs_f_type_t) SMACK_MAGIC:
        return "smackfs";
case (statfs_f_type_t) SMB_SUPER_MAGIC:
        return "smbfs";
case (statfs_f_type_t) SOCKFS_MAGIC:
        return "sockfs";
case (statfs_f_type_t) SQUASHFS_MAGIC:
        return "squashfs";
case (statfs_f_type_t) SYSFS_MAGIC:
        return "sysfs";
case (statfs_f_type_t) TMPFS_MAGIC:
        return "tmpfs";
case (statfs_f_type_t) TRACEFS_MAGIC:
        return "tracefs";
case (statfs_f_type_t) UDF_SUPER_MAGIC:
        return "udf";
case (statfs_f_type_t) USBDEVICE_SUPER_MAGIC:
        return "usbdevfs";
case (statfs_f_type_t) VBOXSF_SUPER_MAGIC:
        return "vboxsf";
case (statfs_f_type_t) MSDOS_SUPER_MAGIC:
        return "vfat";
case (statfs_f_type_t) V9FS_MAGIC:
        return "v9fs";
case (statfs_f_type_t) XENFS_SUPER_MAGIC:
        return "xenfs";
case (statfs_f_type_t) XFS_SUPER_MAGIC:
        return "xfs";
case (statfs_f_type_t) Z3FOLD_MAGIC:
        return "z3fold";
case (statfs_f_type_t) ZONEFS_MAGIC:
        return "zonefs";
case (statfs_f_type_t) ZSMALLOC_MAGIC:
        return "zsmalloc";
