static const char* const af_names[] = { 
        [AF_APPLETALK] = "AF_APPLETALK",
        [AF_IB] = "AF_IB",
        [AF_MCTP] = "AF_MCTP",
        [AF_INET6] = "AF_INET6",
        [AF_NETLINK] = "AF_NETLINK",
        [AF_ROSE] = "AF_ROSE",
        [AF_NETROM] = "AF_NETROM",
        [AF_XDP] = "AF_XDP",
        [AF_TIPC] = "AF_TIPC",
        [AF_WANPIPE] = "AF_WANPIPE",
        [AF_ATMPVC] = "AF_ATMPVC",
        [AF_PHONET] = "AF_PHONET",
        [AF_BRIDGE] = "AF_BRIDGE",
        [AF_CAN] = "AF_CAN",
        [AF_BLUETOOTH] = "AF_BLUETOOTH",
        [AF_SMC] = "AF_SMC",
        [AF_NFC] = "AF_NFC",
        [AF_SECURITY] = "AF_SECURITY",
        [AF_RXRPC] = "AF_RXRPC",
        [AF_ECONET] = "AF_ECONET",
        [AF_AX25] = "AF_AX25",
        [AF_RDS] = "AF_RDS",
        [AF_LLC] = "AF_LLC",
        [AF_KCM] = "AF_KCM",
        [AF_IUCV] = "AF_IUCV",
        [AF_INET] = "AF_INET",
        [AF_X25] = "AF_X25",
        [AF_ATMSVC] = "AF_ATMSVC",
        [AF_PPPOX] = "AF_PPPOX",
        [AF_PACKET] = "AF_PACKET",
        [AF_IEEE802154] = "AF_IEEE802154",
        [AF_IRDA] = "AF_IRDA",
        [AF_MPLS] = "AF_MPLS",
        [AF_KEY] = "AF_KEY",
        [AF_NETBEUI] = "AF_NETBEUI",
        [AF_SNA] = "AF_SNA",
        [AF_VSOCK] = "AF_VSOCK",
        [AF_ALG] = "AF_ALG",
        [AF_QIPCRTR] = "AF_QIPCRTR",
        [AF_ASH] = "AF_ASH",
        [AF_UNIX] = "AF_UNIX",
        [AF_DECnet] = "AF_DECnet",
        [AF_CAIF] = "AF_CAIF",
        [AF_ISDN] = "AF_ISDN",
        [AF_IPX] = "AF_IPX",
};
