static const char* const capability_names[] = { 
        [CAP_PERFMON] = "cap_perfmon",
        [CAP_AUDIT_READ] = "cap_audit_read",
        [CAP_NET_BIND_SERVICE] = "cap_net_bind_service",
        [CAP_LINUX_IMMUTABLE] = "cap_linux_immutable",
        [CAP_SYS_NICE] = "cap_sys_nice",
        [CAP_SYS_RESOURCE] = "cap_sys_resource",
        [CAP_SETGID] = "cap_setgid",
        [CAP_NET_BROADCAST] = "cap_net_broadcast",
        [CAP_SETFCAP] = "cap_setfcap",
        [CAP_AUDIT_CONTROL] = "cap_audit_control",
        [CAP_NET_ADMIN] = "cap_net_admin",
        [CAP_SYS_TIME] = "cap_sys_time",
        [CAP_KILL] = "cap_kill",
        [CAP_FSETID] = "cap_fsetid",
        [CAP_LEASE] = "cap_lease",
        [CAP_SYS_CHROOT] = "cap_sys_chroot",
        [CAP_WAKE_ALARM] = "cap_wake_alarm",
        [CAP_FOWNER] = "cap_fowner",
        [CAP_MAC_ADMIN] = "cap_mac_admin",
        [CAP_SYS_RAWIO] = "cap_sys_rawio",
        [CAP_SYS_BOOT] = "cap_sys_boot",
        [CAP_AUDIT_WRITE] = "cap_audit_write",
        [CAP_DAC_READ_SEARCH] = "cap_dac_read_search",
        [CAP_BLOCK_SUSPEND] = "cap_block_suspend",
        [CAP_CHECKPOINT_RESTORE] = "cap_checkpoint_restore",
        [CAP_SETPCAP] = "cap_setpcap",
        [CAP_BPF] = "cap_bpf",
        [CAP_NET_RAW] = "cap_net_raw",
        [CAP_SYSLOG] = "cap_syslog",
        [CAP_SYS_TTY_CONFIG] = "cap_sys_tty_config",
        [CAP_IPC_OWNER] = "cap_ipc_owner",
        [CAP_SYS_ADMIN] = "cap_sys_admin",
        [CAP_SYS_PTRACE] = "cap_sys_ptrace",
        [CAP_MAC_OVERRIDE] = "cap_mac_override",
        [CAP_SYS_MODULE] = "cap_sys_module",
        [CAP_CHOWN] = "cap_chown",
        [CAP_SYS_PACCT] = "cap_sys_pacct",
        [CAP_IPC_LOCK] = "cap_ipc_lock",
        [CAP_MKNOD] = "cap_mknod",
        [CAP_DAC_OVERRIDE] = "cap_dac_override",
        [CAP_SETUID] = "cap_setuid",
};
